#!/usr/bin/env python
# Copyright (c) 2000-2001 ActiveState Corporation.
# See the file LICENSE.txt for licensing information.

import sys, os, re
from xpcom import components, ServerException, nsError
import logging
log = logging.getLogger('koEnviron')

#---- globals

if sys.platform.startswith("win"):
    _gEnvStrRe = re.compile("^(?P<name>[^&=\r\n\t]+)=(?P<value>.*)$")
    _gEnvStrResToSkip = []
else:
    _gEnvStrRe = re.compile("^(?P<name>[\w]+)=(?P<value>.*)$")
    # On linux we want to skip function definitions like this:
    #
    #    mc=()
    #    {
    #        echo hello;
    #        ls --color=auto ~;
    #        echo bye
    #    }
    # 
    # or this:
    #
    #    mc=() {  mkdir -p $HOME/.mc/tmp 2>/dev/null;
    #     chmod 700 $HOME/.mc/tmp;
    #     MC=$HOME/.mc/tmp/mc-$$;
    #     /usr/bin/mc -P "$@" >"$MC";
    #     cd "`cat $MC`";
    #     rm -i -f "$MC";
    #     unset MC
    #    }
    #
    _gEnvStrResToSkip = [re.compile("^[_\w]+=\(\).*$"),  # function defs
                         re.compile("^\s+.*$"), # leading whitespace == in function def
                         re.compile("^{.*$"), # function start
                         re.compile("^}$"), # function end
                        ]


#---- support routines

def _ParseEnvStr(envStr):
    """Return the (env var name, env var value) pair from the given
    'name=value' environment string.
    """
    # skip unwanted lines formats
    for reToSkip in _gEnvStrResToSkip:
        if reToSkip.search(envStr):
            return None

    # parse out name and value
    envStrMatch = _gEnvStrRe.search(envStr)
    if envStrMatch:
        name = envStrMatch.group("name")
        if sys.platform.startswith("win"):
            name = name.upper()  # put everything upper on Windows
        value = envStrMatch.group("value")
        return name, value
    else:
        log.warn("bogus environment string: '%s'" % envStr)


# simple interface to be able to work with the environment
class KoEnviron:
    _com_interfaces_ = [components.interfaces.koIEnviron]
    _reg_clsid_ = "{68E2A458-186F-4e2c-898E-99F3B21DA819}"
    _reg_contractid_ = "@activestate.com/koEnviron;1"
    _reg_desc_ = "System Environment Interaction Service"

    def has(self, key):
        return os.environ.has_key(key)
    def get(self, key):
        return os.environ[key]
    def set(self, key, value):
        os.environ[key] = value


class KoUserEnviron:
    _com_interfaces_ = [components.interfaces.koIUserEnviron]
    _reg_clsid_ = "{B6289D1A-983E-46e0-85E2-75221D44A5FC}"
    _reg_contractid_ = "@activestate.com/koUserEnviron;1"
    _reg_desc_ = "User Environmnet Interaction Service"

    def __init__(self, startupEnvFileName=None):
        # By default the startup environment file is in the Komodo AppData
        # directory. For testing purposes, though, it can be useful to
        # specify a test file.
        self._startupEnv = None
        if startupEnvFileName is None:
            self._startupEnvBaseName = "startup-env.tmp"
            koDirs = components.classes["@activestate.com/koDirs;1"].\
                getService(components.interfaces.koIDirs)
            self.startupEnvFileName = os.path.join(koDirs.hostUserDataDir,
                                                   "startup-env.tmp")
        else:
            self.startupEnvFileName = startupEnvFileName
    
    def _GetStartupEnv(self):
        if not self._startupEnv:
            self._startupEnv = {}
            # read in startup env delta file and fill in the startup env
            fin = open(self.startupEnvFileName, "r")
            content = fin.read()
            fin.close()
            if not sys.platform.startswith("win"):
                # Replace "\"-continued lines in *Linux* environment dumps.
                # Cannot do this on Windows because a "\" at the end of the
                # line does not imply a continuation.
                content = content.replace("\\\n", "")
            for line in content.split("\n"):
                line = line.rstrip()
                if not line:
                    continue  # skip black lines
                item =  _ParseEnvStr(line)
                if item:
                    self._startupEnv[item[0]] = item[1]

    def has(self, key):
        self._GetStartupEnv()
        if sys.platform.startswith("win"):
            key = key.upper()  # put everything upper on Windows
        return self._startupEnv.has_key(key)

    def get(self, key):
        self._GetStartupEnv()
        if sys.platform.startswith("win"):
            key = key.upper()  # put everything upper on Windows
        return self._startupEnv[key]

    def GetEnvironmentStrings(self):
        self._GetStartupEnv()
        return ["%s=%s" % (item[0], item[1])\
            for item in self._startupEnv.items()]

    def GetEncodedEnvironment(self):
        self._GetStartupEnv()
        envStrings = self.GetEnvironmentStrings()
        return '\n'.join(envStrings)


class KoEnvironUtils:
    _com_interfaces_ = [components.interfaces.koIEnvironUtils]
    _reg_clsid_ = "{cb1e55b4-6837-4659-a9e0-9f6ab4acb54d}"
    _reg_contractid_ = "@activestate.com/koEnvironUtils;1"
    _reg_desc_ = "Environment Variable Utilities"

    def MergeEnvironmentStrings(self, baseEnvStrs, diffEnvStrs):
        envDict = {}
        #print "MergeEnvironmentStrings: merge %s into %s" % (diffEnvStrs,
        #                                                     baseEnvStrs)
        for baseEnvStr in baseEnvStrs:
            item = _ParseEnvStr(baseEnvStr)
            if item:
                envDict[item[0]] = item[1]
        for diffEnvStr in diffEnvStrs:
            #print "merge in '%s' ..." % diffEnvStr
            item = _ParseEnvStr(diffEnvStr)
            if item:
                name = item[0]
                value = item[1]
            else:
                continue
            # an empty value means: remove env. var.
            if value == "":
                if envDict.has_key(name):
                    del envDict[name]
                #print "... delete '%s' variable" % name
            # interpolate any other value into the resultant dictionary
            else:
                value = self._Interpolate(value, envDict)
                envDict[name] = value
                #print "... %s=%s" % (name, value)
        return ["%s=%s" % item for item in envDict.items()]

    def _Interpolate(self, envValue, envDict):
        """Interpolate the variable in the given envDict into the envValue.
        Use platform-specific syntax, i.e. %FOO% on Windows and $FOO or $(FOO)
        elsewhere.
        """
        if sys.platform.startswith("win"):
            envRefRe = re.compile("""
                ^(?P<pre>.*)                    # text before env. var ref
                (?P<envRef>%                    # env. var. reference
                    (?P<envName>[^&=\r\n\t%]+)  # env. var. name
                %)
                (?P<post>.*)$                   # test after env. var ref
                """, re.VERBOSE)
        else:
            envRefRe = re.compile(r"""
                ^(?P<pre>.*)                    # text before env. var ref
                (?P<envRef>
                    \$\b(?P<envName>\w+)\b      # simple $FOO like reference
                    |
                    \$\((?P<envName2>\w+)\)     # parenthesized $(FOO) like ref
                )
                (?P<post>.*)$                   # test after env. var ref
                """, re.VERBOSE)
        interpolated = envValue
        while 1:
            envRefMatch = envRefRe.search(interpolated)
            if envRefMatch:
                envName = envRefMatch.group("envName") \
                          or envRefMatch.group("envName2")
                if sys.platform.startswith("win"):
                    envName = envName.upper()
                # Tough call on how interpolation of undefined variables
                # should be handled because:
                #    windows> echo %NOTDEFINED%
                #    %NOTDEFINED%
                #    linux> echo $NOTDEFINED
                #
                # We will expect the Linux behaviour on both Windows and Linux
                # for now.
                interpolated = envRefMatch.group("pre") \
                               + envDict.get(envName, "") \
                               + envRefMatch.group("post")
            else:
                break
        #print "_Interpolate: interpolate '%s' -> '%s'" % (envValue, interpolated)
        return interpolated

