#!python
# Copyright (c) 2000-2003 ActiveState Corporation.
# See the file LICENSE.txt for licensing information.

"""Services for using templates in Komodo.
Particularly the "New File" dialog.
"""
#TODO:
#  - mechanism to get the "My Templates" stuff into the user folder on
#    a clean first start

import os
import sys
import re
import types
import logging
import copy

from xpcom import components
from xpcom.server import WrapObject, UnwrapObject
from koTreeView import TreeView


log = logging.getLogger("koTemplateService")
#log.setLevel(logging.DEBUG)


class TemplateServiceError(Exception):
    pass


class Node(dict):
    def __init__(self):
        self.files = []
    def isContainer(self):
        return len(self.keys()) > 0
    def dump(self, level=0):
        basenames = [os.path.basename(f['path']) for f in self.files]
        print "files=%s" % basenames
        for name, node in self.items():
            print "%sNode '%s':" % ("  "*level, name),
            node.dump(level+1)


class KoTemplateService:
    _com_interfaces_ = [components.interfaces.koITemplateService]
    _reg_clsid_ = "{9774E3D9-8BAD-4E4D-B759-A89CD03F2A8E}"
    _reg_contractid_ = "@activestate.com/koTemplateService;1"
    _reg_desc_ = "Komodo Template Service"
    
    def __init__(self):
        self.koDirSvc = components.classes["@activestate.com/koDirs;1"].getService()
        self.rootTemplateNode = None
        self.loaded = 0

    def initializeUserTemplateTree(self):
        """Create an empty personal templates tree if there isn't one.
        
        <AppDataDir>/
            templates/
                ReadMe.txt
                My Templates/
                    Sample Custom Template.txt
        """
        try:
            templatesDir = self.getUserTemplatesDir()
            if os.path.exists(templatesDir):
                return
            mytemplatesDir = os.path.join(templatesDir, "My Templates")
            os.makedirs(mytemplatesDir)
            
            readmeTxt = os.path.join(templatesDir, "ReadMe.txt")
            freadme = open(readmeTxt, 'w')
            freadme.write("""
        Komodo User File Templates Directory

Place template files in the "My Templates" folder to have them appear
in Komodo's "New File" dialog.  Your custom templates will appear along
with Komodo's standard set of template files.

To create your own templates files, create a file in Komodo and select
"File | Save As Template..." or simply copy files into the "My Templates"
folder.

See "Custom Templates" in Komodo's on-line help for more information.
""")
            freadme.close()
            
            sampleTxt = os.path.join(mytemplatesDir, "Sample Custom Template.txt")
            fsample = open(sampleTxt, "w")
            fsample.write("""
This is a sample custom template for Komodo. The actual template file is:
    %s

You can create your own templates with any content you wish.
See "Custom Templates" in Komodo's on-line help for more details.
""" % sampleTxt)
            fsample.close()

        except EnvironmentError, ex:
            log.exception(ex)
            _gLastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"]\
                 .getService(components.interfaces.koILastErrorService)
            _gLastErrorSvc.setLastError(ex.errno, str(ex))
            raise ServerException(nsError.NS_ERROR_FAILURE, str(ex))

    def _load_walk(self, dname, node):
        for f in os.listdir(dname):
            path = os.path.join(dname, f)
            if f == ".consign": # Dev build support: skip .consign files.
                continue
            if os.path.isdir(path):
                if f not in node:
                    node[f] = Node()
                self._load_walk(path, node[f])
            elif os.path.isfile(path):
                fdata = {"path": path,
                         "template-name": os.path.splitext(f)[0]}
                node.files.append(fdata)

    def loadTemplates(self):
        """Load template tree from disk."""
        log.info("KoTemplateService.loadTemplates()")
        templateDirs = [
            self.getUserTemplatesDir(),
            self.getDefaultTemplatesDir(),
        ]
        
        # only add the sharedTemplate directory if it exists
        sharedTemplates = self.getSharedTemplatesDir()
        if os.path.exists(sharedTemplates):
            templateDirs.append(sharedTemplates)
        
        self.rootTemplateNode = Node()
        for templateDir in templateDirs:
            self._load_walk(templateDir, self.rootTemplateNode)
        #self.rootTemplateNode.dump()
        self.loaded = 1

    def getTemplateTree(self): # Not in IDL
        """Return a copy of the template tree."""
        if not self.loaded:
            raise TemplateServiceError("You must call .loadTemplates() "
                                       "before .getTemplateTree().")
        return copy.deepcopy(self.rootTemplateNode)

    def dumpTemplates(self):
        """Dump the currently loaded template tree."""
        print "Template Tree:",
        self.rootTemplateNode.dump(1)

    def getDefaultTemplatesDir(self):
        return os.path.join(self.koDirSvc.installDir, "default-templates")

    def getUserTemplatesDir(self):
        return os.path.join(self.koDirSvc.userDataDir, "templates")

    def getSharedTemplatesDir(self):
        return os.path.join(self.koDirSvc.commonDataDir, "templates")


class KoTemplatesView(TreeView):
    _com_interfaces_ = [components.interfaces.koITemplatesView,
                        components.interfaces.nsITreeView]
    _reg_clsid_ = "{53F98D02-237E-417C-A298-D8089AE516D2}"
    _reg_contractid_ = "@activestate.com/koTemplatesView;1"
    _reg_desc_ = "Komodo Templates nsITreeView"

    def __init__(self):
        TreeView.__init__(self) #, debug="templates") #XXX
        self._data = []
        self._tree = None
        self._sortedBy = None

    def setData(self, data):
        # Called by the template categories view when its selection changes.
        self._data = data
        if self.log:
            self.log.debug('setData %r', data)
        self._tree.beginUpdateBatch()
        self._tree.rowCountChanged(0,len(self._data));
        self._tree.endUpdateBatch()
        
    def getSelectedTemplate(self):
        index = self._tree.selection.currentIndex
        return self._data[index]["path"]

    #---- nsITreeView methods
    def get_rowCount(self):
        if self.log:
            self.log.debug("row count %d", len(self._data))
        return len(self._data)

    def getCellText(self, row, col):
        if self.log:
            self.log.debug("getCellText %d:%s", row, col)
        try:
            datum = self._data[row][col]
        except IndexError:
            # Silence this, it is too annoying.
            # c.f. http://bugs.activestate.com/Komodo/show_bug.cgi?id=27487
            #log.error("no %sth result" % row)
            return ""
        except KeyError:
            log.error("unknown template column id: '%s'" % col)
            return ""
        if type(datum) not in (types.StringType, types.UnicodeType):
            datum = str(datum)
        return datum

    def hasNextSibling(self, rowIndex, afterIndex):
        """From the nsITreeView.idl docs:
        
        HasNextSibling is used to determine if the row at rowIndex has a
        nextSibling that occurs *after* the index specified by
        afterIndex.  Code that is forced to march down the view looking
        at levels can optimize the march by starting at afterIndex+1.
        """
        if afterIndex+1 >= len(self._data):
            return 0
        else:
            return 1

    def getImageSrc(self, row, col):
        return "chrome://jaguar/skin/images/newTemplate.png"


class KoTemplateCategoriesView(TreeView):
    _com_interfaces_ = [components.interfaces.koITemplateCategoriesView,
                        components.interfaces.nsITreeView]
    _reg_clsid_ = "{0723776F-EE9D-4F20-B438-3AF893192346}"
    _reg_contractid_ = "@activestate.com/koTemplateCategoriesView;1"
    _reg_desc_ = "Komodo Template Categories nsITreeView"

    def __init__(self):
        TreeView.__init__(self) #, debug="categories") #XXX
        self._tree = None
        self._sortedBy = None
        # The table used to fill in the tree. It is a list of dicts. Each
        # dict represents one row in the tree/outliner. This dictionary is
        # also (because it is convenient) used to store data in addition to
        # the named rows of the XUL tree. e.g.,
        #  [ {"category-name": "My Templates",  # row 0
        #     "node": <Node instance for this category>},
        #    ... ]
        # This attribute is re-generated from self.templateTree whenever
        # the rows change (say, by the user opening a category with
        # sub-categories).
        self._data = []
        
        self._prefSvc = components.classes["@activestate.com/koPrefService;1"]\
                        .getService().prefs # global prefs
        self.templateTree = None # Working copy of template tree Nodes.
        self.selectedTemplateByCategory = None
        self.categoryIsOpen = None

    def initialize(self, templateSvc, templatesView):
        # Need to unwrap these Python XPCOM object because we access
        # methods on them that are not in the IDL.
        self.templateSvc = UnwrapObject(templateSvc)
        self.templatesView = UnwrapObject(templatesView)
        
        if not self.templateSvc.loaded:
            self.templateSvc.loadTemplates()
        self.templateTree = self.templateSvc.getTemplateTree() # a working copy
        
        # Restore the user selections from prefs
        stbcStr = self._prefSvc.getStringPref("templates_selected_template_by_category")
        try:
            self.selectedTemplateByCategory = eval(stbcStr)
        except SyntaxError, ex:
            self.selectedTemplateByCategory = {}
        ocStr = self._prefSvc.getStringPref("templates_open_categories")
        try:
            self.categoryIsOpen = eval(ocStr)
        except SyntaxError, ex:
            self.categoryIsOpen = {}

        self._updateRowData()

    def saveSelections(self):
        # Persist the user selections in prefs.
        row = self.selection.currentIndex
        self._prefSvc.setStringPref("templates_selected_category",
                                    self._data[row]["category-name"])
        self._prefSvc.setStringPref("templates_selected_template_by_category",
                                    repr(self.selectedTemplateByCategory))
        self._prefSvc.setStringPref("templates_open_categories",
                                    repr(self.categoryIsOpen))

    def templateSelectionChanged(self):
        tindex = self.templatesView.selection.currentIndex
        index = self.selection.currentIndex
        if index < 0 or len(self._data) <= index:
            return
        cdict = self._data[self.selection.currentIndex]
        categoryPath = cdict["category-path"]
        try:
            templateName = cdict["node"].files[tindex]["template-name"]
        except IndexError: # if no template for this category, tindex==0
            pass
        else:
            if self.log:
                self.log.debug("remember user last selected '%s' template for '%s' "
                      "category", templateName, categoryPath)
            self.selectedTemplateByCategory[categoryPath] = templateName
    

    def getDefaultCategoryIndex(self):
        selectedCategory = self._prefSvc.getStringPref("templates_selected_category")
        for index in range(len(self._data)):
            if self._data[index]["category-name"] == selectedCategory:
                break
        else:
            index = 0 # default to first row
        if self.log:
            self.log.debug("getDefaultCategoryIndex: '%s' (index %s)",
                  selectedCategory, index)
        return index

    def getDefaultTemplateIndex(self):
        index = self.selection.currentIndex
        if index < 0 or len(self._data) <= index:
            return 0
        cdict = self._data[self.selection.currentIndex]
        categoryPath = cdict["category-path"]
        try:
            templateName = self.selectedTemplateByCategory[categoryPath]
        except KeyError:
            templateName = ""
            index = 0 # default to first row
        else:
            files = cdict["node"].files
            for index in range(len(files)):
                if files[index]["template-name"] == templateName:
                    break
            else:
                index = 0 # default to first row
        # Note: index==0 seems to work to select nothing if there are no
        # templates for this category.
        if self.log:
            self.log.debug("getDefaultTemplateIndex: '%s' (index %s, category '%s')",
                  templateName, index, categoryPath)
        return index

    def _getCategoryRows(self, rows, node, level=0, parentIndex=-1, path=[]):
        subnames = node.keys()
        subnames.sort()
        for subname in subnames:
            subnode = node[subname]
            subpath = list(path) # get a copy
            subpath.append(subname)
            categoryPath = '/'.join(subpath)
            cdict = {"category-name": subname,
                     "category-path": categoryPath,
                     "node": subnode,
                     "level": level,
                     "parent-index": parentIndex}
            rows.append(cdict)
            if self.categoryIsOpen.get(categoryPath, 0):
                self._getCategoryRows(rows, subnode, level=level+1,
                                      parentIndex=len(rows)-1, path=subpath)

    def _updateRowData(self):
        """Update self._data for changes in the XUL tree rows or new source
        data.
        """
        if not self.templateTree:
            raise TemplateServiceError("Cannot get category rows before "
                                       "loading from disk.")
        rows = []
        self._getCategoryRows(rows, self.templateTree)
        self._data = rows
        self._tree.beginUpdateBatch()
        self._tree.rowCountChanged(0,len(self._data));
        self._tree.endUpdateBatch()

    #---- nsITreeView methods
    def get_rowCount(self):
        if self.log:
            self.log.debug("rowCount %d", len(self._data))
        return len(self._data)

    def getCellText(self, row, col):
        if self.log:
            self.log.debug("getCellText %d:%s", row, col)
        try:
            datum = self._data[row][col]
        except IndexError:
            log.error("no %sth find result" % row)
            return ""
        except KeyError:
            log.error("unknown category column id: '%s'" % col)
            return ""
        if type(datum) not in (types.StringType, types.UnicodeType):
            datum = str(datum)
        return datum

    def isContainer(self, row):
        if self.log:
            self.log.debug("isContainer %d", row)
        return self._data[row]["node"].isContainer()

    def isContainerEmpty(self, row):
        return 0

    def isContainerOpen(self, row):
        if self.log:
            self.log.debug("isContainerOpen %d", row)
        categoryPath = self._data[row]["category-path"]
        return self.categoryIsOpen.get(categoryPath, 0)

    def toggleOpenState(self, row):
        if self.log:
            self.log.debug("toggleOpenState %d", row)
        categoryPath = self._data[row]["category-path"]
        isOpen = self.categoryIsOpen.get(categoryPath, 0)
        self.categoryIsOpen[categoryPath] = not isOpen

        # Update row data.
        self._updateRowData()
        self.selection.currentIndex = row
        self.selection.select(row);        

    def getParentIndex(self, row):
        if self.log:
            self.log.debug("getParentIndex %d", self._data[row]["parent-index"])
        return self._data[row]["parent-index"]

    def getLevel(self, row):
        if self.log:
            self.log.debug("getLevel %d", self._data[row]["level"])
        return self._data[row]["level"]

    def selectionChanged(self):
        if self.log:
            self.log.debug("selectionChanged")
        index = self.selection.currentIndex
        if index < 0 or len(self._data) <= index:
            return
        cdict = self._data[index]
        self.templatesView.setData(cdict["node"].files)

    def getImageSrc(self, row, col):
        categoryPath = self._data[row]["category-path"]
        if self.categoryIsOpen.get(categoryPath, 0):
            return "chrome://jaguar/skin/images/folder-open.png"
        else:
            return "chrome://jaguar/skin/images/folder-closed.png"
    
    def hasNextSibling(self, rowIndex, afterIndex):
        """From the nsITreeView.idl docs:
        
        HasNextSibling is used to determine if the row at rowIndex has a
        nextSibling that occurs *after* the index specified by
        afterIndex.  Code that is forced to march down the view looking
        at levels can optimize the march by starting at afterIndex+1.
        """
        if afterIndex+1 >= len(self._data):
            return 0
        else:
            return (self._data[rowIndex]["parent-index"]
                    == self._data[afterIndex+1]["parent-index"])


if __name__ == "__main__":
    templateSvc = components.classes["@activestate.com/koTemplateService;1"].getService()
    templateSvc.loadTemplates()
    templateSvc.dumpTemplates()

