#!python
# Copyright (c) 2004 ActiveState Corp, a division of Sophos Plc.
# See the file LICENSE.txt for licensing information.

"""Get and manage Code Intelligence data about source code of many languages.

The Code Intelligence system is one for generating and managing code
structure information on given source code. See the spec for more details:
    http://specs.activestate.com/Komodo_3.0/func/code_intelligence.html
Usage examples follow.

Scan a file and return data in CIX format:

    import codeintel
    cix = codeintel.scan("foo.py", "Python")

General program structure for a runtime C.I. system:

    # Initialize the manager at application startup.
    manager = codeintel.Manager()
    manager.initialize()

    # Request scans on certain files.
    # - Scans are performed asynchronously and data is loaded into the
    #   database (CIDB).
    r = codeintel.ScanRequest(os.path.join("play", "sample.py"),
                              "Python",
                              codeintel.PRIORITY_OPEN)
    manager.addRequest(r)

    #XXX Describe how to get data out, both synchronously and asynchronously
    #    (by getting a notification when requested scans are completed).

    # Finalize the manager when shutting down app.
    manager.finalize()
"""
#TODO:
# - test processing of Blackfile.py (it breaks: CIX is not wellformed), then
#   the database is locked on subsequent attempts:
#
#        Exception in thread Code Intel Scheduler:
#        Traceback (most recent call last):
#          File "C:\Python22\Lib\threading.py", line 408, in __bootstrap
#            self.run()
#          File "D:\trentm\as\Apps\Komodo-devel\install\release\INSTALLDIR\python\Lib\sit
#        e-packages\codeintel\__init__.py", line 896, in run
#            (request.getCanonicalPath(), request.md5sum))
#          File "C:\Python22\Lib\site-packages\sqlite\main.py", line 255, in execute
#            self.rs = self.con.db.execute(SQL % parms)
#        OperationalError: database is locked
#
#   then bad things happen: scheduling is hung
# - start in on the code browser and autocomplete/calltip UIs; fleshing this
#   out will make it more clear what sort of information the database needs
#   to hold
# - implement a command-line interface to the database for development,
#   testing, and browsing (less important with "SQLite Database Browser" to
#   browse a database).
# - spec out the database schema more (at least those parts relevant to
#   Python) and get Shane involved in reviewing this
# - get people started on (or prep for it at least) the other language
#   engines; this might be a bit premature
# - XML support
# - refine, refine, refine
#
# Dev Notes on PySQLite and Multi-threading:
# - There seem to be some caveats when using multi-threading (particularly
#   on Un*x) with PySQLite/SQLite. I have submitted one crash report to
#   pysqlite-users: avoid usage of Connection.db.sqlite_busy_handler().
#   I think multi-threading will work fine, but we should be able to move
#   to all database modification coming from one thread if necessary: have
#   a results Queue and the scheduler thread can feed those results into the
#   database; rather than the current situation where each _ScanRunner thread
#   does the db updating. c.f. the THREADSAFE flags for building PySQLite
#   and/or SQLite.
# - We _do_ have to properly handle SQL_BUSY retvals -> OperationalError
#   exceptions in the Python binding.

import os
import sys
import time
import md5
import pprint
import threading
import stat
import types

import logging
import sqlite
import process

from codeintel.common import *
from codeintel.scheduler import Scheduler, BatchScheduler
from codeintel.database import Database
from codeintel.engines import LanguageEngineRegistry
from codeintel import citdl


#---- global variables

_version_ = (0, 1, 0)
log = logging.getLogger("codeintel.__init__")
#log.setLevel(logging.DEBUG)



#---- public convenience methods

def scan(filename, language, md5sum=None, mtime=None):
    """Scan the given filename and return code intelligence data (CIX).
    
        "filename" is the path to a file to scan.
        "language" is a string indicating what kind of content the file
            holds. This is used to get an appropriate Language Engine Driver
            from the LanguageEngineRegistry. Generally the proper string
            is just the name of the language, e.g., "Python", "Perl".
        "md5sum" (optional) if the MD5 hexdigest has already been calculated
            for the content, it can be passed in here. Otherwise this
            is calculated.
        "mtime" (optional) is a modified time for the file (in seconds since
            the "epoch"). If it is not specified the file is stat'd.

    If there are no language engines for the given content type a
    CodeIntelError is raised.
    """
    log.debug("scan(filename='%s', language='%s', md5sum=%r, mtime=%r)",
              filename, language, md5sum, mtime)
    if mtime is None:
        mtime = int(os.stat(filename)[stat.ST_MTIME])
    fin = open(filename, "r")
    try:
        s = fin.read()
    finally:
        fin.close()
    return scanString(s, language, path=filename, md5sum=md5sum, mtime=mtime)


def scanString(s, language, path="<string>", md5sum=None, mtime=None):
    """Scan the given filename and return code intelligence data (CIX).
    
        "s" is the content to scan.
        "language" is a string indicating what kind of content the file
            holds. This is used to get an appropriate Language Engine Driver
            from the LanguageEngineRegistry. Generally the proper string
            is just the name of the language, e.g., "Python", "Perl".
        "path" (optional) is the path from which the content originates.
            This path appears in the resulting XML so may be relevant. If
            not specified "<string>" is used.
        "md5sum" (optional) if the MD5 hexdigest has already been calculated
            for the content, it can be passed in here. Otherwise this
            is calculated.
        "mtime" (optional) is a modified time for the file (in seconds since
            the "epoch"). If it is not specified the _current_ time is used.
            Note that the default is not to stat() the file and use that
            because the given content might not reflect the saved file state.

    If there are no language engines for the given content type a
    CodeIntelError is raised.
    """
    if md5sum is None: md5sum = md5.new(content).hexdigest()
    if mtime is None: mtime = int(time.time())
        
    registry = LanguageEngineRegistry()
    drivers = registry.getDrivers(language)
    if not drivers:
        raise CodeIntelError("There are no Code Intelligence Language "
                             "Engines register for '%s' content"
                             % language)
    mergedCIX = ""
    for driver in drivers:
        r = ScanRequest(path, language, PRIORITY_IMMEDIATE,
                        content=s, md5sum=md5sum, mtime=mtime)
        cix = driver.scan(r)
        if not mergedCIX:
            mergedCIX = cix
        else:
            raise NotImplementedError("Can't yet merge CIX documents.")
    return mergedCIX



#---- internal support stuff

def _getPyClassMembers(obj):
    # Recurse into base classes looking for attributes
    members = []
    try:
        members = members + dir(obj)
        for b in obj.__bases__:
            for m in _getPyClassMembers(b):
                if m not in members:
                    members.append(m)
    except AttributeError:
        pass
    return members


def _findPyClassConstructor(classObj):
    # Given a class object, return a function object used for the
    # constructor (ie, __init__() ) or None if we can't find one.
    try:
        return classObj.__init__.im_func
    except AttributeError:
        for base in classObj.__bases__:
            rc = _findPyClassConstructor(base)
            if rc is not None:
                return rc
    return None


def _getPyObjCallTip(obj):
    """Return a suitable call tip for the given Python object."""
    if obj is None:
        return ""

    name = ""
    argStr = ""
    doc = None
    argOffset = 0
    if type(obj) == types.ClassType:
        # Look for the highest __init__ in the class chain.
        funcObj = _findPyClassConstructor(obj)
        if funcObj is None:
            funcObj = lambda: None
        else:
            argOffset = 1
        name = "%s.%s" % (obj.__name__, funcObj.__name__)
    elif type(obj) == types.MethodType:
        # bit of a hack for methods - turn it into a function
        # but we drop the "self" param.
        funcObj = obj.im_func
        argOffset = 1
        name = funcObj.__name__
    else:
        funcObj = obj
        try:
            name = funcObj.__name__
        except AttributeError:
            pass
    # Try and build one for Python defined functions
    if type(funcObj) in (types.FunctionType, types.LambdaType):
        try:
            realArgs = funcObj.func_code.co_varnames[argOffset:funcObj.func_code.co_argcount]
            defaults = funcObj.func_defaults or []
            defaults = list(map(lambda name: "=%s" % name, defaults))
            defaults = [""] * (len(realArgs)-len(defaults)) + defaults
            items = map(lambda arg, dflt: arg+dflt, realArgs, defaults)
            if funcObj.func_code.co_flags & 0x4:
                items.append("...")
            if funcObj.func_code.co_flags & 0x8:
                items.append("***")
            argStr = ", ".join(items)
            argStr = "(%s)" % argStr
        except:
            pass
    # See if we can use the docstring.
    if hasattr(obj, "__doc__") and obj.__doc__:
        pos = obj.__doc__.find("\n")
        if pos<0 or pos>70: pos=70
        doc = obj.__doc__[:pos]
    # Put it together.
    calltip = name+argStr
    if doc:
        if calltip: calltip += "\n"  # calltip can be empty for, e.g., str(
        calltip += doc
    return calltip



#---- public Manager class

class Manager:
    """Manage a long-running interaction with the Code Intelligence system.
    
    This class can be sub-classed to customize some handling. The following
    methods are useful for overriding:
        requestCompleted(request)
    """
    MIN_CIDB_VERSION = (0, 14)  # minimum supported database version

    def __init__(self, registry=None):
        """Create a Code Intel Manager.
        
            "registry" (optional) can be used to specify a customized
                LanguageEngineRegistry instance to use for mapping languages
                to CILEs (Code Intel Language Engines). If not specified the
                default set of mappings is used.
        """
        if registry is None:
            self._cileRegistry = LanguageEngineRegistry()
        else:
            self._cileRegistry = registry
        self._importHandlers = {}
        self._scheduler = None # the scheduler thread, started as required
        self._dbPath = None # the path to the SQLite database file
        self._connections = {} # thread name -> cached connection
        self._cx = None
        
        # Used for batch updates.
        self._batchScheduler = None
        self._lastBatchUpdateErrors = None
        self._batchProgressCache = None
        self._batchUpdateCompletedEvent = threading.Event()

    def initialize(self, dbPath=None):
        """Initialize the Code Intelligence system.
        
            "dbPath" (optional) is the full path to the SQLite database
                to use. If not specified the default, "./codeintel.db",
                is used.
        
        If the database path does not exist when this is called then a new
        database is created. A CodeIntelError, DatabaseError, or
        EnvironmentError may be raised if the database path cannot be read or
        written to. The database is wiped if it is unsupported (e.g.,
        insufficient version, incorrect platform).
        """
        if not dbPath:
            self._dbPath = os.path.abspath("codeintel.db")
        else:
            self._dbPath = dbPath

        # For now we just trash the existing database if it is too old or
        # whatever. Eventually we may want to provide a richer mechanism:
        # - error handling in a separate (overridable) method with default of
        #   trashing the database
        # - think about accommodating a UI so that a status dialog can be put
        #   up while a conversion is done
        if os.path.isfile(self._dbPath):
            try:
                cx = self.getDatabaseConnection()
                try:
                    # Ensure we support this database version.
                    log.info("open CIDB at '%s'" % self._dbPath)
                    db = Database(self, cx)
                    if db.getVersionInfo() < self.MIN_CIDB_VERSION:
                        raise CodeIntelError(
                            "database '%s' version, %s, is unsupported"
                            % (self._dbPath, db.getVersion()))
                    import socket
                    if db.getHostname() != socket.gethostname():
                        raise CodeIntelError(
                            "database '%s' hostname, '%s', does not match "
                            "the current hostname: '%s'"
                            % (self._dbPath, db.getHostname(),
                               socket.gethostname()))
                    if db.getPlatform() != getPlatform():
                        raise CodeIntelError(
                            "database '%s' platform, '%s', does not match "
                            "the current platform: '%s'"
                            % (self._dbPath, db.getPlatform(),
                               getPlatform()))
                finally:
                    cx.close()
            except CodeIntelError, ex:
                log.warn("%s: removing database" % ex)
                self._removeDatabase()

        if not os.path.isfile(self._dbPath):
            self._createNewDatabase()

        self._cx = self.getDatabaseConnection()

    def _removeDatabase(self):
        try:
            os.remove(self._dbPath)
        except EnvironmentError, ex:
            # Try a little harder.
            time.sleep(1)
            try:
                os.chmod(self._dbPath, 0777)
                os.remove(self._dbPath)
                if os.path.exists(self._dbPath):
                    raise EnvironmentError("File remains after attempted "
                                           "removal: '%s'" % self._dbPath)
            except EnvironmentError, ex2:
                raise CodeIntelError("Could not remove database: it must "
                                     "be removed manually: %s" % ex2)

    def _createNewDatabase(self):
        # Create a new database.
        log.info("create new CIDB at '%s'" % self._dbPath)
        cx = self.getDatabaseConnection()
        db = Database(self, cx)
        try:
            db.create()
        finally:
            cx.close()


    def finalize(self):
        if self._batchScheduler:
            self._batchScheduler.stop()
        if self._scheduler:
            self._scheduler.stop()
        if self._cx:
            self._cx.close()

    def _startScheduler(self):
        if self._dbPath is None:
            raise CodeIntelError("must initialize the Manager before "
                                 "starting the scheduler")
        self._scheduler = Scheduler(Scheduler.MODE_DAEMON, self,
                                    self.requestCompleted)
        self._scheduler.start()

    def getDatabaseConnection(self):
        """Return a connection to the database.
        
        This returns a connection object conforming to Python's DB API 2.0
        spec:
            http://www.python.org/peps/pep-0249.html
        """
        #XXX Look into encoding issues and the use of the 'encoding'
        #    argument.
        if 0 and log.isEnabledFor(logging.DEBUG):
            command_logfile = sys.stdout
        else:
            command_logfile = None
        cx = sqlite.connect(self._dbPath, timeout=3000,
                            command_logfile=command_logfile)
        return cx

    def getCILEDrivers(self, language):
        """Return all registered Language Engine Drivers for this language.

        Returns a list of LanguageEngineDriver instances. If there are no
        engines for this langage an empty list is returned.
        """
        return self._cileRegistry.getDrivers(language)

    def getImportHandler(self, language):
        """Return an "import"-handler for the given language.
        
        Returns None if don't know how to handle imports for this language.
        Each language-specific ImportHandler object is a singleton.
        """
        if language not in self._importHandlers:
            handler = citdl.getImportHandler(language)
            self._importHandlers[language] = handler
        return self._importHandlers[language]

    def isSupportedLanguage(self, language):
        """Return true iff the given language is supported."""
        return language in self._cileRegistry.getLanguages()

    def getSupportedLanguages(self):
        """Return a list of the supported languages."""
        return self._cileRegistry.getLanguages()

    def addRequest(self, request):
        """Make a scan request.
        
            "request" is a ScanRequest instance.
        """
        if self._scheduler is None:  # lazily create the scheduler thread
            self._startScheduler()
        self._scheduler.addRequest(request)

    def stageRequest(self, request):
        """Stage a scan request.
        
            "request" is a ScanRequest instance.
        
        Staging a scan request involves putting it "on deck" for a certain
        period (called the "staging delay") before actually queuing it.
        Subsequently staged requests for the same file within that period
        replace the earlier request on deck. The point is to reduce
        unnecessary scanning for a rapidly changing file.
        """
        if self._scheduler is None:  # lazily create the scheduler thread
            self._startScheduler()
        self._scheduler.stageRequest(request)

    def requestCompleted(self, request):
        """Called when a scan request is successfully completed.
        
            "request" is the ScanRequest instance.
        
        This is made available for sub-classes to override. By default this
        method does nothing. Note: This is called in the sub-thread used to
        complete that request.
        """
        pass

    def batchUpdateRequest(self, type, path, language=None, extra=None):
        """Queue up a request for a particular batch update of the db.
        
            "type" is one of "language" (meaning a language installation),
                "cix" (meaning import a CIX file), or "directory" (meaning
                scan a source directory).
            "path" is the full path to a language install, CIX file, or
                source directory, depending on the type.
            "language" is the language of the language installation, None
                for type=="cix", and the language of the library dir.
            "extra" is an optional (null if not used) extra value depending
                on the type, path and/or language of the request that may be
                request for processing it. For example, a PHP language batch
                update request uses the "extra" field to specify the
                "php.ini"-config-file path.

        Note: this cannot be called between bacthUpdateStart and
        batchUpdateCompleted().
        """
        if self._batchScheduler is None:
            self._batchScheduler = BatchScheduler(self,
                                                  self.batchUpdateProgress,
                                                  self._batchUpdateCompleted,
                                                  self.batchUpdateError)
        self._batchScheduler.addRequest(type, path, language, extra)

    def batchUpdateStart(self):
        # Setup thread(s) to handle the batch requests and return. (I.e. this
        # is asynchronous.
        # Put the main scheduler on hold.
        if self._scheduler is not None:
            self._scheduler.pause()
        self._batchUpdateCompletedEvent.clear()
        self._lastBatchUpdateErrors = []
        self._batchProgressCache = {}
        self._batchScheduler.start()

    def batchUpdateStop(self):
        """Stop the current batch update."""
        self._batchScheduler.stop()

    def batchUpdateWait(self):
        """Convenience method to wait for batch update completion."""
        self._batchUpdateCompletedEvent.wait()

    def batchUpdateGetErrors(self):
        """Return a list of errors from the last batch update."""
        return self._lastBatchUpdateErrors

    def _batchUpdateCompleted(self, reason):
        if self._scheduler is not None:
            self._scheduler.resume()
        self._lastBatchUpdateErrors = self._batchScheduler.errors
        self._batchScheduler = None
        self.batchUpdateCompleted(reason)
        self._batchUpdateCompletedEvent.set()

    _SHOW_PROGRESS_ON_ONE_LINE = True
    def batchUpdateProgress(self, stage, obj):
        """Called when progress is made on a set of batch update requests.
        
            "stage" is a string defining the current processing stage.
            "obj" is some object relevant to a particular stage to describe
                is state of progress.
        
        Subclasses can override this for custom handling.
        """
        cache = self._batchProgressCache or {}
        if stage == "Preparing database": # a.k.a. removing db indices
            name, current, total = obj
            line = "%s: %s" % (stage, name)
        elif stage == "Restoring indices":
            name, current, total = obj
            line = "%s: %s" % (stage, name)
        elif stage == "Gathering files":
            line = "Gathering files: %s files" % obj
        elif stage == "Scanning": # 'obj' is a request object
            # Scanning: [#####        ] (1/12) ...tel\xsltcile.py ETA: 5 min
            remaining = self._batchScheduler.getNumFilesToProcess()+1
            total = cache.setdefault("total_files", remaining)
            # Progress meter section
            percent = float(total-remaining)/float(total)
            METER_WIDTH = 20
            meterTemplate = "%%-%ds" % METER_WIDTH
            meter = meterTemplate % ("#"*int(METER_WIDTH*percent))
            # Count and file section
            # 79: display width; 25: other stuff, e.g. "Scanning", "ETA: ..."
            file = obj.path
            COUNT_AND_FILE_WIDTH = 79 - 25 - METER_WIDTH
            count = "(%d/%d)" % (total-remaining+1, total)
            FILE_WIDTH = COUNT_AND_FILE_WIDTH - len(count) - 1
            if len(file) > FILE_WIDTH:
                file = "..."+file[-FILE_WIDTH+3:]
            countAndFile = "%s %s" % (count, file)
            # ETA section
            now = time.time()
            if "starttime" in cache:
                elapsedsecs = now - cache["starttime"]
                totalsecs = elapsedsecs / percent
                remainsecs = totalsecs - elapsedsecs
                if remainsecs < 60.0:
                    eta = "ETA: %d sec" % int(remainsecs)
                elif (remainsecs/60.0 < 60.0):
                    eta = "ETA: %d min" % int(remainsecs/60.0)
                else:
                    eta = "ETA: %d hr" % int(remainsecs/3600.0)
            else:
                cache["starttime"] = now
                eta = ""
            # Put sections together
            line_template = "Scanning: [%%s] %%-%ds %%s" % COUNT_AND_FILE_WIDTH
            line = line_template % (meter, countAndFile, eta)
        else:
            line = "%s: %s" % (stage, obj)
        if self._SHOW_PROGRESS_ON_ONE_LINE and "last_line" in cache:
            length = len(cache["last_line"])
            sys.stdout.write( "\b \b"*length )
        sys.stdout.write(line)
        if not self._SHOW_PROGRESS_ON_ONE_LINE:
            sys.stdout.write('\n')
        cache["last_line"] = line
        self._batchProgressCache = cache

    def batchUpdateCompleted(self, reason):
        """Called when a set of batch update requests are completed.
        
        Subclasses can override this for custom handling.
        """
        cache = self._batchProgressCache or {}
        if "starttime" in cache:
            totalsecs = time.time() - cache["starttime"]
            if totalsecs < 60.0:
                elapsed = "%d seconds" % int(totalsecs)
            elif (totalsecs/60.0 < 60.0):
                elapsed = "%d minute(s)" % int(totalsecs/60.0)
            else:
                elapsed = "%d hour(s)" % int(totalsecs/3600.0)
        else:
            elapsed = None
        errors = self._lastBatchUpdateErrors
        if self._SHOW_PROGRESS_ON_ONE_LINE and "last_line" in cache:
            length = len(cache["last_line"])
            sys.stdout.write( "\b \b"*length )
        if reason == "completed":
            sys.stdout.write("Batch update completed.")
        elif reason == "stopped":
            sys.stdout.write("Batch update was cancelled.")
        elif reason == "error":
            sys.stdout.write("Batch update errored out.")
        else:
            sys.stdout.write("Batch update completed (%s)." % reason)
        if elapsed: sys.stdout.write(" Running time: %s." % elapsed)
        if errors: sys.stdout.write(" There were %d errors:" % len(errors))
        sys.stdout.write("\n")
        if errors:
            print "\t"+"\n\t".join(errors)

    def batchUpdateError(self, errmsg):
        """Called when there is an error in batch update processing.
        
            "errmsg" is a short string describing the error.
        
        Note that all errors can be retrieved after a batch update is
        complete by calling batchUpdateGetErrors()

        Subclasses can override this for custom handling.
        """
        print "Batch update error: %s" % errmsg

    def _parseCITDLExpr(self, expr):
        """Parse a CITDL expression into <symbol ref> and <scope> parts."""
        #XXX This should really be in citdl.py.
        return expr.split(None, 1)

    def _getMembersPyEvalFallback(self, symbolRef):
        """See if the symbol reference matches one of the modules in the
        current process run-time. ...to ensure being no worse that previous
        AutoComplete implementations in Komodo.
        
        Returns None if nothing suitable was found.
        """
        try:
            namespace = sys.modules.copy()
            #NOTE: the PythonWin equivalent code also has:
            #      - namespace.update(__main__.__dict__); and
            #      - Adds the current context from its Interactive Shell.
            #      Do we want to add those?
            obj = eval(symbolRef, namespace)
        except:
            return None
        members = None
        if obj is not None:
            members = []

            try:
                members = members + dir(obj)
            except AttributeError:
                pass # object has no __dict__
            if hasattr(obj, "__class__"):
                members = members + _getPyClassMembers(obj.__class__)
            members = [m for m in members if m[0] != '_']

            # The object may be a COM object with typelib support - lets see
            # if we can get its props. (contributed by Stefan Migowsky)
            try:
                # Get the automation attributes
                ms = obj.__class__._prop_map_get_.keys()
                # See if there is an write only property 
                # could be optimized
                for m in obj.__class__._prop_map_put_.keys():
                    if m not in ms:
                        ms.append(m)
                # append to the already evaluated list
                members = members + ms
            except AttributeError:
                pass
            
            if not members: members = None
        return members

    def _getMembersFallback(self, expr, content, line, language):
        """A dumb fallback mechanism for finding members.
        
            "expr" is the CITDL expression.
            "content" is the file content.
            "line" is the line in the content on which AutoComplete was
                triggered.
            "language" is the file's language.
        
        Basic algorithms from PythonWin. Returns a list of found members
        (the empty list if none found).
        """
        symbolRef = self._parseCITDLExpr(expr)[0]
        members = None
        if language == "Python" and '(' not in symbolRef:
            #XXX Technically eval'ing any "symbolRef" is potentially
            #    dangerous. __getattr__ code *will* be run. We specifically
            #    disallow eval'ing expressions with "()" because those are
            #    more likely to have a side-effect.
            members = self._getMembersPyEvalFallback(symbolRef)
        if members is None:
            # Heuristics a-la "complete word": The idea is to find other
            # usages of the current binding and assume, that it refers to the
            # same object (or at least, to an object of the same type)
            # Contributed to PythonWin by Vadim Chugunov [vadimch@yahoo.com]

            #NOTE: PythonWin limits the search to the current class using
            #      its class browser. Perhaps we should do that here with
            #      our scope information?

            if language == "Python":
                pattern = re.compile(r"\b%s\.(\w+)" % re.escape(symbolRef))
            else:
                raise CodeIntelError("don't know how to make a suitable "
                                     "dumb-get-members pattern for '%s'"
                                     % language)
            try:
                members = pattern.findall(content)
            except re.error:
                # With the added "re.escape()" (when this code was adapted
                # for use in the CodeIntel system) this should raise, but
                # let's be paranoid.
                members = []
            unique = {}
            for m in members:
                unique[m] = 1
            members = unique.keys()

            # Only filter out __*__ members and not _* members because
            # local private and protected attributes on a class should be
            # included.
            members = [m for m in members
                       if not (m[:2]=="__" and m[-2:]=="__")]

            # Ignore the word currently to the right of the dot - probably a
            # red-herring.
            lines = content.splitlines()
            herrings = re.findall(pattern, lines[line-1])
            members = [m for m in members if m not in herrings]
        return members or []

    def getMembers(self, language, path, line, expr,
                   scopeTable=None, scopeId=None, content=None):
        """Return the likely members for the given expression and scope.

            "language" is a language name (e.g., "Python", "Perl") of the
                working file. This can be relevant for some things, e.g.
                import semantics.
            "path" and "line" together indicate the scope from which to
                interpret the expression. Line is 1-based.
            "expr" is a CITDL expression identifying the object/symbol/type
                for which to return members.
            "scopeTable" and "scopeId" (optional) can be passed it to
                express the scope more accurately than "path" and "line" --
                if, say, the caller is adjusting for recent edits not in the
                database.
            "content" (optional) is the file content. This may be used, if
                passed in, for fallback "dumb" completion handling.

        Raises a CodeIntelError if there is a problem processing.
        """
        #start = time.clock() #XXX
        cu = self._cx.cursor()
        try:
            context = citdl.Context(self, language, path, line)
            try:
                members = context.getMembers(cu, expr, scopeTable, scopeId)
            except CodeIntelError:
                if content:
                    members = self._getMembersFallback(expr, content,
                                                       line, language)
                    if members:
                        return members
                raise
            if not members and content:
                #XXX Typically we should trust citdl.Context's evaluation
                #    but we don't. Try "self.cx." in Database.create, for
                #    example. In any case, as long as determining members
                #    is not perfect it is better to have false positives
                #    than nothing.
                members = self._getMembersFallback(expr, content,
                                                   line, language)
            return members
        finally:
            cu.close()
            #print "XXX getMembers took: %f seconds" % (time.clock() - start)

    def _getCallTipsPyEvalFallback(self, symbolRef):
        """See if the symbol reference matches one of the modules in the
        current process run-time. ...to ensure being no worse that previous
        AutoComplete implementations in Komodo.
        
        Returns None if nothing suitable was found.
        """
        try:
            namespace = sys.modules.copy()
            #NOTE: the PythonWin equivalent code also has:
            #      - namespace.update(__main__.__dict__); and
            #      - Adds the current context from its Interactive Shell.
            #      Do we want to add those?
            obj = eval(symbolRef, namespace)
        except:
            return None
        calltips = None
        if obj is not None:
            calltip = _getPyObjCallTip(obj)
            if calltip: calltips = [calltip]
        return calltips

    def _getCallTipsFallback(self, expr, content, line, language):
        """A dumb fallback mechanism for finding calltips.
        
            "expr" is the CITDL expression.
            "content" is the file content.
            "line" is the line in the content on which a CallTip was
                triggered.
            "language" is the file's language.
        
        Basic algorithms from PythonWin. Returns a list of found CallTips
        (the empty list if none found).
        """
        symbolRef = self._parseCITDLExpr(expr)[0]
        calltips = None
        if language == "Python" and '(' not in symbolRef:
            #XXX Technically eval'ing any "symbolRef" is potentially
            #    dangerous. __getattr__ code *will* be run. We specifically
            #    disallow eval'ing expressions with "()" because those are
            #    more likely to have a side-effect.
            calltips = self._getCallTipsPyEvalFallback(symbolRef)
        return calltips or []

    def getCallTips(self, language, path, line, expr,
                    scopeTable=None, scopeId=None, content=None):
        """Return the likely calltips for the given expression and scope.

            "language" is a language name (e.g., "Python", "Perl") of the
                working file. This can be relevant for some things, e.g.
                import semantics.
            "path" and "line" together indicate the scope from which to
                interpret the expression. Line is 1-based.
            "expr" is a CITDL expression identifying the object/symbol/type
                for which to return calltips.
            "scopeTable" and "scopeId" (optional) can be passed it to
                express the scope more accurately than "path" and "line" --
                if, say, the caller is adjusting for recent edits not in the
                database.
            "content" (optional) is the file content. This may be used, if
                passed in, for fallback "dumb" completion handling.

        Typically there is just one calltip, but there _can_ be many: for
        overloaded methods in C++ for example, for ambiguous symbol references
        in dynamic languages. Raises a CodeIntelError if there is a problem
        processing.
        """
        #start = time.time() #XXX
        cu = self._cx.cursor()
        try:
            context = citdl.Context(self, language, path, line)
            try:
                calltips = context.getCallTips(cu, expr, scopeTable, scopeId)
            except CodeIntelError:
                if content:
                    calltips = self._getCallTipsFallback(expr, content,
                                                         line, language)
                    if calltips:
                        return calltips
                raise
            if not calltips and content:
                calltips = self._getCallTipsFallback(expr, content,
                                                     line, language)
            return calltips
        finally:
            cu.close()
            #print "XXX getCallTips took: %f seconds" % (time.time() - start)

    def getScopeForFileAndLine(self, path, line):
        """This is a rawer form of getScopeInfoForFileAndLine.
        
        Returns (<type>, <id>, <row>) where
            <type> is "module" or "symbol";
            <id> is a CIDB row id into that table; and
            <row> is the row data.
        Returns (None, 0, None) if can't determine.
        """
        cu = self._cx.cursor()
        try:
            try:
                scope = citdl.getScopeForFileAndLine(cu, path, line)
            except CodeIntelError, ex:
                return (None, 0, None)
            else:
                return (scope.table, scope.id, scope.row)
        finally:
            cu.close()

    def getScopeInfoForFileAndLine(self, path, line, language):
        """Return info for presenting the current scope is a UI.
        
            "path" is the full (absolute) path to the file.
            "line" is a 1-based line number.
            "language" is the language of the file.
            
        Returns: (<typeName>, <name>, <image url>, <description>)
        
        If the line is at the global level then the return value is
            ("module", <module name>, None, None)
        If the scope cannot be determined the return value is
            (None, None, None, None)
        """
        from codeintel.citdl import M_NAME, S_TYPE, S_ID, S_SCOPE, S_NAME, S_ATTRIBUTES
        cu = self._cx.cursor()
        try:
            try:
                scope = citdl.getScopeForFileAndLine(cu, path, line)
            except CodeIntelError, ex:
                #XXX Commonly get the following. Need to look into it.
                #       WARN: codeintel.__init__: could not find file in CIDB: 'D:\trentm\as\Apps\Komodo-devel\Blackfile.py'
                #    Silence the warning for now though.
                #log.warn(str(ex))
                if 1: #XXX
                    import traceback
                    traceback.print_exception(*sys.exc_info())
                retval = (None, None, None, None)
            else:
                if scope.table == "module":
                    retval = ("module", scope.row[M_NAME], None, None)
                else:
                    row = scope.row
                    import cb
                    type, id, scope, name, attrStr =\
                        row[S_TYPE], row[S_ID], row[S_SCOPE], row[S_NAME], row[S_ATTRIBUTES]
                    attributes = parseAttributes(attrStr)
                    typeName = symbolType2Name(type)
                    imageURL = cb.getImageURLForSymbol(typeName, attributes)
                    desc = cb.getDescForSymbol(typeName, name, attributes,
                                               scope, language)
                    retval = (typeName, name, imageURL, desc)
            return retval
        finally:
            cu.close()



#---- mainline/self-test

def test_PriorityQueue():
    import pythoncom
    
    class Foo(object):
        def __init__(self, name):
            self.id = str(pythoncom.CreateGuid())
            self.name = name
        def __repr__(self):
            return "<Foo '%s' (%s)>" % (self.name, self.id)
    queue = _PriorityQueue(0)
    queue.put( (2, time.time(), Foo("second")) )
    time.sleep(0.1)
    queue.put( (1, time.time(), Foo("zzz")) )
    time.sleep(0.1)
    queue.put( (1, time.time(), Foo("first")) )
    time.sleep(0.1)
    queue.put( (1, time.time(), Foo("asdf")) )
    time.sleep(0.1)
    queue.put( (1, time.time(), Foo("aqer")) )
    time.sleep(0.1)
    queue.put( (1, time.time(), Foo("xzcv")) )
    time.sleep(0.1)
    t = Foo("third")
    queue.put( (3, time.time(), t) )
    time.sleep(0.1)
    queue.put( (1, time.time(), Foo("awer")) )
    time.sleep(0.1)
    queue.put( (1, time.time(), Foo("opiu")) )
    time.sleep(0.1)
    queue.put( (1, time.time(), Foo("asdf")) )
    time.sleep(0.1)
    queue.put( (1, time.time(), Foo("qwer")) )
    time.sleep(0.1)
    queue.put( (1, time.time(), t) )
    time.sleep(0.1)
    for i in range(5):
        print queue.get()
    queue.put( (0, time.time(), Foo("sentinel")) )
    while queue.queue:
        print queue.get()


def test_ScanRequestQueue():
    queue = _ScanRequestQueue(0)
    r1 = ScanRequest("foo.py", "Python", PRIORITY_IMMEDIATE)
    r1.id = 1
    queue.put( (r1.priority, time.time(), r1) )
    time.sleep(0.1)
    r2 = ScanRequest("bar.py", "Python", PRIORITY_IMMEDIATE)
    r2.id = 2
    queue.put( (r2.priority, time.time(), r2) )
    time.sleep(0.1)
    # This new request should replace the earlier (and higher priority) r1
    # and, hence, get onto the queue ahead of r2.
    r3 = ScanRequest("foo.py", "Python", PRIORITY_BACKGROUND)
    r3.id = 3
    queue.put( (r3.priority, time.time(), r3) )

    p, t, r = queue.get()
    assert r.id == r3.id, "Expected r3 to get first on the queue. That didn't happen."
    assert len(queue.queue) == 1, "Expected only one item (r2) to remain on queue."
    print "test_ScanRequestQueue... ok"

def test_StagingScanRequestQueue():
    sys.stdout.write("test_StagingScanRequestQueue...")
    queue = _StagingScanRequestQueue(0, 1.0)
    # With a staging delay of one second only the last ScanRequest for
    # "foo.py" should get remain on the queue, and then only after the second
    # has elapsed.
    for i in range(5):
        r = ScanRequest("foo.py", "Python", PRIORITY_IMMEDIATE)
        r.id = i
        #sys.stdout.write("[stage %d]" % i)
        queue.stage( (r.priority, time.time(), r) )
        time.sleep(0.1)

    # Cheat a little bit and poke into the queue's internal data structure
    # to test that none of these are on the queue yet.
    #sys.stdout.write("[assert non staged]")
    assert len(queue.queue) == 0, "Expected no requests to be on queue yet."

    # After at least one "staging delay" there should be just one request
    # on the queue and it should be that last one (id=4)
    #sys.stdout.write("[wait for staging]")
    time.sleep(2.0)
    assert (len(queue.queue) == 1,
            "Expected only one request for foo.py to be on the queue by now: "
            "len(queue.queue)==%d" % len(queue.queue))
    #sys.stdout.write("[get item]")
    p, t, r = queue.get()
    assert r.id==4 and r.path=="foo.py", "Unexpected request on the queue: id=%d, path=%r"

    queue.finalize()
    print "ok"

def _play():
    print "_play..."
    dbpath = os.path.abspath("codeintel.db")
    if os.path.exists(dbpath):
        # Start with a fresh database everytime.
        os.remove(dbpath)

    class MyManager(Manager):
        def requestCompleted(self, request):
            print "completed scan of '%s' (%s)"\
                  % (request.path, request.language)

    m = MyManager()
    m.initialize()
    try:
        hiPy = os.path.join("play", "hi.py")
        files = [
            (os.path.join("play", "raven.py"), "Python"),
            (os.path.join("play", "assign.py"), "Python"),
            (hiPy, "Python"),
            (os.path.join("play", "sample.py"), "Python"),
            (os.path.join("test", "scan_inputs", "php5_sample.php"), "PHP"),
            (os.path.join("test", "test_eval_citdl", "hi.py"), "Python"),
        ]
        for file, lang in files:
            print "request scan of '%s' (%s)" % (file, lang)
            r = ScanRequest(file, lang, PRIORITY_OPEN, force=1)
            m.addRequest(r)

        ## Cheat and test pause/resume of the Scheduler
        #print "_play: pausing Scheduler..."
        #m._scheduler.pause()
        #time.sleep(6)
        #print "_play: ...resuming Scheduler"
        #m._scheduler.resume()

        print "_play: start batch update..."
        #m.batchUpdateRequest("language", r"C:\Python22\python.exe",
        #                     "Python")
        #m.batchUpdateRequest("directory", r"D:\applications\Python-2.2\Lib",
        #                     "Python")
        m.batchUpdateRequest("directory", r"D:\trentm\as\Apps\Komodo-devel\src\codeintel",
                             "*")
        m.batchUpdateStart()
        if 1:
            m.batchUpdateWait()
        else:
            print "_play: ...cancelling batch update in 25 seconds..."
            time.sleep(25)
            m.batchUpdateStop()
        print "_play: ...done batch update"

        #pythonCIX = os.path.join("codeintel", "python.cix")
        #print "_play: import '%s'..." % pythonCIX
        #m.batchUpdateRequest("cix", pythonCIX)
        #m.batchUpdateStart()
        #m.batchUpdateWait()
        #errors = m.batchUpdateGetErrors()
        #if errors:
        #    print "Errors importing '%s':\n\t%s" % (pythonCIX, "\n\t".join(errors))
        #print "_play: ...done importing '%s'" % pythonCIX

        ## Request the scan of hi.py again, it should get dropped based
        ## on mtime.
        #time.sleep(1)
        #print "request scan of '%s' (Python) again" % hiPy
        #r = ScanRequest(hiPy, "Python", PRIORITY_OPEN)
        #m.addRequest(r)

        #time.sleep(1)
        #fin = open(hiPy, 'r')
        #content = fin.read()
        #fin.close()
        #content += "\n\ndef poke(): pass # added this method\n"
        #r = ScanRequest(hiPy, "Python", PRIORITY_OPEN, content=content)
        #m.addRequest(r)
    
        while m._scheduler and m._scheduler.getNumRequests(): # let scheduler finish its scans
            time.sleep(1)
        time.sleep(5)
    
        ## Test that looking in the built-in scope works.
        #expr = "__version__"; path = hiPy; line = 5
        #print "get members for '%s' at %s:%d (to test built-in scope)"\
        #      % (expr, path, line)
        #members = m.getMembers("Python", os.path.abspath(path), line, expr)
        #print "members are:", members

        ## Dump the Code Browser
        #cx = m.getDatabaseConnection()
        #cu = cx.cursor()
        #try:
        #    print "\n\n--------- Code Browser ---------"
        #    import cb
        #    root = cb.CBRootNode()
        #    for file, lang in files:
        #        root.addPath(cu, os.path.abspath(file))
        #    root.pprint()
        #    print "-----------------------------------"
        #finally:
        #    cu.close()
        #    cx.close()

        ## Try out some of the info retrieval routines.
        #ravenPy = os.path.abspath(os.path.join("play", "raven.py"))
        #print
        #for line in (1, 12, 14, 17, 18, 19, 33):
        #    retval = m.getScopeInfoForFileAndLine(ravenPy, line)
        #    print "getScopeInfoForFileAndLine('.../raven.py', line=%r): %r"\
        #          % (line, retval)
        #print
        #triggerPoints = [
        #    (31, "Raven"),
        #    (29, "raven"),
        #    (15, "quoth"),
        #    (16, "quoth"),
        #    (30, "random"),
        #    (30, "hi"),
        #]
        #for line, citdl in triggerPoints:
        #    try:
        #        retval = m.getMembers("Python", ravenPy, line, citdl)
        #    except CodeIntelError, ex:
        #        retval = "error: %s" % ex
        #    print "getMembers('.../raven.py', line=%r, citdl=%r): %s\n"\
        #          % (line, citdl, retval)
        #print
    finally:
        m.finalize()
        print "done"

def _test_bug_32305():
    print "_test_bug_32305..."
    REGEN = 1
    dbpath = os.path.abspath("codeintel.db")
    if REGEN:
        if os.path.exists(dbpath):
            # Start with a fresh database everytime.
            os.remove(dbpath)

    m = Manager()
    m.initialize()
    try:
        if REGEN:
            pythonCIX = os.path.join("codeintel", "python.cix")
            print "_play: import '%s'..." % pythonCIX
            m.batchUpdateRequest("cix", pythonCIX)
            m.batchUpdateStart()
            m.batchUpdateWait()
            errors = m.batchUpdateGetErrors()
            if errors:
                print "Errors importing '%s':\n\t%s" % (pythonCIX, "\n\t".join(errors))
            print "_play: ...done importing '%s'" % pythonCIX
    
            files = [r"C:\Python22\Lib\os.py"]
            for file in files:
                print "request scan of '%s' (Python)" % file
                r = ScanRequest(file, "Python", PRIORITY_OPEN)
                m.addRequest(r)
    
            while m._scheduler and m._scheduler.getNumRequests(): # let scheduler finish its scans
                time.sleep(1)
            time.sleep(5)
    

        # Test that looking in the 'os' module works as expected.
        fooPy = "foo.py"
        tests = [
            #{"content": "import os, _winreg\n\n",
            # "exprs": ["os", "os.path"]},
            #{"content": "from os import path",
            # "exprs": ["path"]},
            #{"content": "from os import *\n\n",
            # "exprs": ["path"]},
            #{"content": "from os.path import *\n\n",
            # "exprs": ["abspath"]},
            {"content": "import re\n\n",
             "exprs": ["re"]},
        ]
        
        for test in tests:
            print "request scan of '%s' (Python)" % fooPy
            r = ScanRequest(fooPy, "Python", PRIORITY_OPEN,
                            content=test["content"], force=1)
            m.addRequest(r)
            time.sleep(1)

            for expr in test["exprs"]:
                path = fooPy; line = 3
                print "\nget members for '%s' at %s:%d (to test built-in scope)"\
                      % (expr, path, line)
                try:
                    members = m.getMembers("Python", os.path.abspath(path), line, expr)
                    members.sort()
                    print "\nmembers are:", members
                except CodeIntelError, ex:
                    log.error(str(ex))
    finally:
        m.finalize()
        print "done"

if __name__ == "__main__":
    __file__ = os.path.abspath(sys.argv[0])
    logging.basicConfig()
    #test_PriorityQueue()
    #test_ScanRequestQueue()
    #test_StagingScanRequestQueue()
    _test_bug_32305()
    #_play()

